"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.VM = exports.VMState = exports.Debugger = void 0;
const context_1 = require("./context");
const instruction_1 = require("./byte-compiler/instruction");
const default_1 = require("./types/default");
const stack_1 = require("./utils/stack");
const error_1 = require("./utils/error");
const with_intrinsics_1 = require("./types/with-intrinsics");
const boolean_1 = require("./types/boolean");
const function_1 = require("./types/function");
const evaluation_1 = require("./vm/evaluation");
const number_1 = require("./types/number");
const map_1 = require("./types/map");
const list_1 = require("./types/list");
const set_immediate_1 = require("./utils/set-immediate");
const string_1 = require("./types/string");
const events_1 = __importDefault(require("events"));
const object_value_1 = require("./utils/object-value");
const call_1 = require("./vm/call");
const limited_stack_1 = require("./utils/limited-stack");
class Debugger {
    constructor() {
        this.breakpoint = false;
        this.nextStep = false;
    }
    debug(...segments) {
        console.debug(...segments);
        return default_1.DefaultType.Void;
    }
    setBreakpoint(breakpoint) {
        this.breakpoint = breakpoint;
        return this;
    }
    getBreakpoint(vm) {
        return this.breakpoint;
    }
    next() {
        this.nextStep = true;
        return this;
    }
    resume() {
        if (!this.breakpoint) {
            return Promise.resolve();
        }
        return new Promise((resolve) => {
            const check = () => {
                if (!this.breakpoint) {
                    resolve();
                }
                else if (this.nextStep) {
                    this.nextStep = false;
                    resolve();
                }
                else {
                    (0, set_immediate_1.setImmediate)(check);
                }
            };
            check();
        });
    }
    interact(vm, _source) {
        const me = this;
        console.warn('Debugger is not setup.');
        console.info(vm);
        me.breakpoint = false;
    }
}
exports.Debugger = Debugger;
var VMState;
(function (VMState) {
    VMState[VMState["PREPARATION"] = 0] = "PREPARATION";
    VMState[VMState["PENDING"] = 1] = "PENDING";
    VMState[VMState["FINISHED"] = 2] = "FINISHED";
    VMState[VMState["STOPPED"] = 3] = "STOPPED";
})(VMState = exports.VMState || (exports.VMState = {}));
class VM {
    constructor(options) {
        var _a, _b, _c, _d;
        this.ACTIONS_PER_LOOP = 80000;
        this.MAX_FRAMES = 10000;
        this.STACK_LIMIT = 512;
        this.stack = new Array(this.STACK_LIMIT);
        this.signal = new events_1.default();
        this.state = VMState.PREPARATION;
        this.maxActionsPerLoop = (_a = options.maxActionsPerLoop) !== null && _a !== void 0 ? _a : this.ACTIONS_PER_LOOP;
        this.actionCount = 0;
        this.sp = 0;
        this.time = -1;
        this.target = options.target;
        this.frames = new limited_stack_1.LimitedStack(this.MAX_FRAMES, options.globals);
        this.contextTypeIntrinsics = options.contextTypeIntrinsics;
        this.handler = options.handler;
        this.debugger = options.debugger;
        this.environmentVariables = (_b = options.environmentVariables) !== null && _b !== void 0 ? _b : new Map();
        this.imports = (_c = options.imports) !== null && _c !== void 0 ? _c : new Map();
        this.externalFrames = (_d = options.externalFrames) !== null && _d !== void 0 ? _d : new stack_1.Stack();
    }
    getTime() {
        return this.time;
    }
    isPending() {
        return this.state === VMState.PENDING;
    }
    getStacktrace() {
        const externalFrames = this.externalFrames.values();
        const frames = this.frames.values();
        const stacktrace = [];
        for (let index = 0; index < externalFrames.length; index++) {
            stacktrace.push(externalFrames[index].getCurrentInstruction());
        }
        for (let index = 0; index < frames.length; index++) {
            stacktrace.unshift(frames[index].getCurrentInstruction());
        }
        return stacktrace;
    }
    setMaxActionsPerLoop(actions) {
        this.maxActionsPerLoop = actions;
        return this;
    }
    getOpenHandles() {
        return this.sp;
    }
    getSignal() {
        return this.signal;
    }
    exit() {
        this.state = VMState.STOPPED;
        this.signal.emit('exit');
    }
    getFrame() {
        return this.frames.peek();
    }
    getFrames() {
        return this.frames;
    }
    pushStack(value) {
        this.stack[this.sp++] = value;
    }
    popStack() {
        return this.stack[--this.sp];
    }
    createFrame(options) {
        const ctx = this.getFrame().fork({
            code: options.code,
            type: context_1.ContextType.Function,
            self: options.self,
            super: options.super,
            outer: options.outer
        });
        this.frames.push(ctx);
        return ctx;
    }
    popFrame() {
        if (this.frames.length === 1) {
            return null;
        }
        const frame = this.frames.pop();
        frame.cp = null;
        return frame;
    }
    exec() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.state = VMState.PENDING;
                this.time = Date.now();
                this.resume((err) => {
                    if (err) {
                        reject(err);
                        return;
                    }
                    resolve();
                });
            });
        });
    }
    resume(done) {
        var _a, _b, _c, _d, _e;
        try {
            while (true) {
                if (!this.isPending()) {
                    done();
                    return;
                }
                const frame = this.getFrame();
                if (frame.code.length === frame.ip) {
                    this.popFrame();
                    continue;
                }
                const instruction = frame.code[frame.ip++];
                frame.cp = instruction.ip;
                switch (instruction.op) {
                    case instruction_1.OpCode.NOOP: {
                        break;
                    }
                    case instruction_1.OpCode.PUSH: {
                        const pushInstruction = instruction;
                        this.pushStack(pushInstruction.value);
                        break;
                    }
                    case instruction_1.OpCode.POP: {
                        this.popStack();
                        break;
                    }
                    case instruction_1.OpCode.GET_GLOBALS: {
                        this.pushStack(frame.globals.scope);
                        break;
                    }
                    case instruction_1.OpCode.GET_OUTER: {
                        this.pushStack(((_a = frame.outer) !== null && _a !== void 0 ? _a : frame.globals).scope);
                        break;
                    }
                    case instruction_1.OpCode.GET_LOCALS: {
                        this.pushStack(frame.locals.scope);
                        break;
                    }
                    case instruction_1.OpCode.GET_SELF: {
                        this.pushStack(frame.locals.get(string_1.Self, this.contextTypeIntrinsics));
                        break;
                    }
                    case instruction_1.OpCode.GET_SUPER: {
                        this.pushStack(frame.locals.get(string_1.Super, this.contextTypeIntrinsics));
                        break;
                    }
                    case instruction_1.OpCode.IMPORT: {
                        const importInstruction = instruction;
                        const code = this.imports.get(importInstruction.path);
                        this.createFrame({ code });
                        break;
                    }
                    case instruction_1.OpCode.ASSIGN: {
                        const value = this.popStack();
                        const key = this.popStack();
                        const base = this.popStack();
                        if (!(base instanceof with_intrinsics_1.CustomValueWithIntrinsics)) {
                            throw new error_1.RuntimeError('Invalid left value in assignment!', this);
                        }
                        base.set(key, value);
                        break;
                    }
                    case instruction_1.OpCode.FUNCTION_DEFINITION: {
                        const functionInstruction = instruction;
                        const fn = new function_1.CustomFunction('anonymous', functionInstruction.code, functionInstruction.arguments, functionInstruction.ignoreOuter ? frame.globals : frame);
                        this.pushStack(fn);
                        break;
                    }
                    case instruction_1.OpCode.CALL: {
                        const callInstruction = instruction;
                        const args = new Array(callInstruction.length);
                        for (let i = 0; i < callInstruction.length; i++) {
                            args[i] = this.popStack();
                        }
                        const fn = this.popStack();
                        if (fn instanceof function_1.CustomFunction) {
                            if (callInstruction.length > fn.arguments.length) {
                                throw new error_1.RuntimeError('Too many arguments.', this);
                            }
                            const newFrame = this.getFrame().fork({
                                code: fn.value,
                                type: context_1.ContextType.Function,
                                outer: fn.outer
                            });
                            (0, call_1.call)(newFrame, fn, args);
                            this.frames.push(newFrame);
                        }
                        break;
                    }
                    case instruction_1.OpCode.CALL_WITH_CONTEXT: {
                        const callInstruction = instruction;
                        const args = new Array(callInstruction.length);
                        for (let i = 0; i < callInstruction.length; i++) {
                            args[i] = this.popStack();
                        }
                        const propertyName = this.popStack();
                        const context = this.popStack();
                        const ret = context.getWithOrigin(propertyName, this.contextTypeIntrinsics);
                        const fn = ret.value;
                        if (fn instanceof function_1.CustomFunction) {
                            const newFrame = this.getFrame().fork({
                                type: context_1.ContextType.Function,
                                code: fn.value,
                                self: context,
                                super: ret.origin instanceof map_1.CustomMap ? ((_b = ret.origin.getIsa()) !== null && _b !== void 0 ? _b : default_1.DefaultType.Void) : null,
                                outer: fn.outer
                            });
                            (0, call_1.callWithContext)(newFrame, fn, args);
                            this.frames.push(newFrame);
                        }
                        break;
                    }
                    case instruction_1.OpCode.CALL_SUPER_PROPERTY: {
                        const callInstruction = instruction;
                        const args = new Array(callInstruction.length);
                        for (let i = 0; i < callInstruction.length; i++) {
                            args[i] = this.popStack();
                        }
                        const property = this.popStack();
                        const context = frame.locals.get(string_1.Super, this.contextTypeIntrinsics);
                        const ret = context.getWithOrigin(property, this.contextTypeIntrinsics);
                        const fn = ret.value;
                        if (fn instanceof function_1.CustomFunction) {
                            const newFrame = this.getFrame().fork({
                                type: context_1.ContextType.Function,
                                code: fn.value,
                                self: frame.self,
                                super: ret.origin instanceof map_1.CustomMap ? ((_c = ret.origin.getIsa()) !== null && _c !== void 0 ? _c : default_1.DefaultType.Void) : null,
                                outer: fn.outer
                            });
                            (0, call_1.callWithContext)(newFrame, fn, args);
                            this.frames.push(newFrame);
                        }
                        break;
                    }
                    case instruction_1.OpCode.CONSTRUCT_MAP: {
                        const mapConstructInstruction = instruction;
                        const map = [];
                        for (let index = 0; index < mapConstructInstruction.length; index++) {
                            const value = this.popStack();
                            const key = this.popStack();
                            map.unshift([key, value]);
                        }
                        this.pushStack(new map_1.CustomMap(new object_value_1.ObjectValue(map)));
                        break;
                    }
                    case instruction_1.OpCode.CONSTRUCT_LIST: {
                        const listConstructInstruction = instruction;
                        const list = [];
                        for (let index = 0; index < listConstructInstruction.length; index++) {
                            const value = this.popStack();
                            list.unshift(value);
                        }
                        this.pushStack(new list_1.CustomList(list));
                        break;
                    }
                    case instruction_1.OpCode.GOTO_A_IF_FALSE: {
                        const condition = this.popStack();
                        if (condition.toTruthy()) {
                            break;
                        }
                        const gotoAInstruction = instruction;
                        frame.ip = gotoAInstruction.goto.ip;
                        break;
                    }
                    case instruction_1.OpCode.GOTO_A_IF_FALSE_AND_PUSH: {
                        const condition = this.popStack();
                        const value = condition instanceof number_1.CustomNumber ? (0, evaluation_1.absClamp01)(condition.value) : +condition.toTruthy();
                        this.pushStack(new number_1.CustomNumber(value));
                        if (value >= 1) {
                            break;
                        }
                        const gotoAInstruction = instruction;
                        frame.ip = gotoAInstruction.goto.ip;
                        break;
                    }
                    case instruction_1.OpCode.GOTO_A_IF_TRUE_AND_PUSH: {
                        const condition = this.popStack();
                        const value = condition instanceof number_1.CustomNumber ? (0, evaluation_1.absClamp01)(condition.value) : +condition.toTruthy();
                        this.pushStack(new number_1.CustomNumber(value));
                        if (value < 1) {
                            break;
                        }
                        const gotoAInstruction = instruction;
                        frame.ip = gotoAInstruction.goto.ip;
                        break;
                    }
                    case instruction_1.OpCode.GOTO_A: {
                        const gotoAInstruction = instruction;
                        frame.ip = gotoAInstruction.goto.ip;
                        break;
                    }
                    case instruction_1.OpCode.GOTO_A_IF_FALSE: {
                        const condition = this.popStack();
                        if (condition.toTruthy()) {
                            break;
                        }
                        const gotoAInstruction = instruction;
                        frame.ip = gotoAInstruction.goto.ip;
                        break;
                    }
                    case instruction_1.OpCode.PUSH_ITERATOR: {
                        const value = this.popStack();
                        const iterator = value[Symbol.iterator]();
                        frame.iterators.push(iterator);
                        break;
                    }
                    case instruction_1.OpCode.POP_ITERATOR: {
                        frame.iterators.pop();
                        break;
                    }
                    case instruction_1.OpCode.NEXT: {
                        const nextInstruction = instruction;
                        let idx = frame.get(nextInstruction.idxVariable, this.contextTypeIntrinsics).toNumber();
                        const iterator = frame.iterators.peek();
                        iterator.index = ++idx;
                        const iteratorResult = iterator.next();
                        this.pushStack(new boolean_1.CustomBoolean(!iteratorResult.done));
                        if (!iteratorResult.done) {
                            frame.set(nextInstruction.variable, iteratorResult.value);
                            frame.set(nextInstruction.idxVariable, new number_1.CustomNumber(idx));
                        }
                        break;
                    }
                    case instruction_1.OpCode.GET_VARIABLE: {
                        const getVariableInstroduction = instruction;
                        const ret = frame.locals.get(getVariableInstroduction.property, this.contextTypeIntrinsics);
                        if (ret instanceof function_1.CustomFunction && getVariableInstroduction.invoke) {
                            const newFrame = this.getFrame().fork({
                                code: ret.value,
                                type: context_1.ContextType.Function,
                                outer: ret.outer
                            });
                            (0, call_1.call)(newFrame, ret, []);
                            this.frames.push(newFrame);
                            break;
                        }
                        this.pushStack(ret);
                        break;
                    }
                    case instruction_1.OpCode.GET_PROPERTY: {
                        const getPropertyInstruction = instruction;
                        const property = this.popStack();
                        const context = this.popStack();
                        if (!(context instanceof with_intrinsics_1.CustomValueWithIntrinsics)) {
                            throw new error_1.RuntimeError(`Unknown path ${property.toString()}.`, this);
                        }
                        const ret = context.getWithOrigin(property, this.contextTypeIntrinsics);
                        if (ret.value instanceof function_1.CustomFunction && getPropertyInstruction.invoke) {
                            const newFrame = this.getFrame().fork({
                                type: context_1.ContextType.Function,
                                code: ret.value.value,
                                self: context,
                                super: ret.origin instanceof map_1.CustomMap ? ((_d = ret.origin.getIsa()) !== null && _d !== void 0 ? _d : default_1.DefaultType.Void) : null,
                                outer: ret.value.outer
                            });
                            (0, call_1.callWithContext)(newFrame, ret.value, []);
                            this.frames.push(newFrame);
                            break;
                        }
                        this.pushStack(ret.value);
                        break;
                    }
                    case instruction_1.OpCode.GET_SUPER_PROPERTY: {
                        const getPropertyInstruction = instruction;
                        const property = this.popStack();
                        const context = frame.locals.get(string_1.Super, this.contextTypeIntrinsics);
                        const ret = context.getWithOrigin(property, this.contextTypeIntrinsics);
                        if (ret.value instanceof function_1.CustomFunction && getPropertyInstruction.invoke) {
                            const newFrame = this.getFrame().fork({
                                type: context_1.ContextType.Function,
                                code: ret.value.value,
                                self: frame.self,
                                super: ret.origin instanceof map_1.CustomMap ? ((_e = ret.origin.getIsa()) !== null && _e !== void 0 ? _e : default_1.DefaultType.Void) : null,
                                outer: ret.value.outer
                            });
                            (0, call_1.callWithContext)(newFrame, ret.value, []);
                            this.frames.push(newFrame);
                            break;
                        }
                        this.pushStack(ret.value);
                        break;
                    }
                    case instruction_1.OpCode.CALL_INTERNAL: {
                        const callInstruction = instruction;
                        const args = new Map();
                        const callback = callInstruction.callback;
                        for (const arg of callInstruction.arguments) {
                            const value = frame.scope.value.get(arg.name);
                            args.set(arg.name.toString(), value);
                        }
                        const immediateRet = callback(this, frame.self, args);
                        // Due to transformations it can happen that return values may not be PromiseLike
                        if (immediateRet === null || immediateRet === void 0 ? void 0 : immediateRet.then) {
                            immediateRet.then((ret) => {
                                this.pushStack(ret);
                                this.resume(done);
                            }).catch(done);
                            return;
                        }
                        this.pushStack(immediateRet);
                        break;
                    }
                    case instruction_1.OpCode.SLICE: {
                        const b = this.popStack();
                        const a = this.popStack();
                        const value = this.popStack();
                        if (value instanceof list_1.CustomList || value instanceof string_1.CustomString) {
                            this.pushStack(value.slice(a, b));
                        }
                        else {
                            this.pushStack(default_1.DefaultType.Void);
                        }
                        break;
                    }
                    case instruction_1.OpCode.NEW: {
                        const value = this.popStack();
                        if (value instanceof map_1.CustomMap) {
                            this.pushStack(value.createInstance());
                        }
                        else {
                            this.pushStack(value);
                        }
                        break;
                    }
                    case instruction_1.OpCode.NEGATE: {
                        const value = this.popStack();
                        this.pushStack(new number_1.CustomNumber(-value.toNumber()));
                        break;
                    }
                    case instruction_1.OpCode.FALSIFY: {
                        const value = this.popStack();
                        this.pushStack(new boolean_1.CustomBoolean(!value.toTruthy()));
                        break;
                    }
                    case instruction_1.OpCode.ISA: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack(new boolean_1.CustomBoolean(a.instanceOf(b, this.contextTypeIntrinsics)));
                        break;
                    }
                    case instruction_1.OpCode.BITWISE_AND: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalBitwiseAnd)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.BITWISE_OR: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalBitwiseOr)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.BITWISE_LEFT_SHIFT: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalBitwiseLeftShift)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.BITWISE_RIGHT_SHIFT: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalBitwiseRightShift)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.BITWISE_UNSIGNED_RIGHT_SHIFT: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalBitwiseUnsignedRightShift)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.ADD: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalAdd)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.SUB: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalSub)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.MUL: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalMul)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.DIV: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalDiv)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.MOD: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalMod)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.POW: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalPow)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.EQUAL: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalEqual)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.NOT_EQUAL: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalNotEqual)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.GREATER_THAN: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalGreaterThan)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.GREATER_THAN_OR_EQUAL: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalGreaterThanOrEqual)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.LESS_THAN: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalLessThan)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.LESS_THAN_OR_EQUAL: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalLessThanOrEqual)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.AND: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalAnd)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.OR: {
                        const b = this.popStack();
                        const a = this.popStack();
                        this.pushStack((0, evaluation_1.evalOr)(a, b));
                        break;
                    }
                    case instruction_1.OpCode.RETURN: {
                        const value = this.popStack();
                        frame.iterators.clear();
                        this.popFrame();
                        this.pushStack(value !== null && value !== void 0 ? value : default_1.DefaultType.Void);
                        break;
                    }
                    case instruction_1.OpCode.GET_ENVAR: {
                        const key = this.popStack();
                        const value = this.environmentVariables.get(key.toString());
                        this.pushStack(new string_1.CustomString(value));
                        break;
                    }
                    case instruction_1.OpCode.BREAKPOINT: {
                        if (this.debugger.getBreakpoint(this)) {
                            this.debugger.interact(this, instruction.source);
                            this.debugger.resume().then(() => {
                                this.resume(done);
                            }).catch(done);
                            return;
                        }
                        break;
                    }
                    case instruction_1.OpCode.BREAKPOINT_ENABLE: {
                        this.debugger.setBreakpoint(true);
                        break;
                    }
                    case instruction_1.OpCode.HALT: {
                        this.state = VMState.FINISHED;
                        this.signal.emit('done');
                        done();
                        return;
                    }
                }
                if (this.actionCount++ === this.maxActionsPerLoop) {
                    this.actionCount = 0;
                    (0, set_immediate_1.setImmediate)(() => {
                        this.resume(done);
                    });
                    return;
                }
            }
        }
        catch (err) {
            done(err);
        }
    }
}
exports.VM = VM;
